
(defvar ceta-command-alist nil
  "Command alist of external command Line.
Key is string shown at minibuffer menu.
Value is list which contains executable program fullpath, 
command line Macro, and some bool options.")

(setq ceta-command-alist
	  '(
		("skelton"
		 ("command-file" . "/usr/X11R6/bin/rxvt")
		 ("command-line" . "%fc -e %lnt")
		 ("command-err" . nil))
		("Sample"
		 ("command-file" . "/local/bin/gcc")
		 ("command-line" . "%fp -S -o %flt.asm %lft")
		 ("command-err" . t))
		("gcc"
		 ("command-file" . "gcc")
		 ("command-line" . "%fc -o %lot %lft")
		 ("command-err" . t))
		("Vi"
		 ("command-file" . "/usr/bin/vi")
		 ("command-line" . "/usr/X11R6/bin/rxvt -e %fc %lft")
		 ("command-err" . nil))
	   ("Akipic(16F84)-Write"
		("command-file" . "/usr/local/bin/akipic")
		("command-line" . "%fp -d 16f84 -w %lft")
		("command-err" . t))
	   ("GPASM(16F84)"
		("command-file" . "/usr/local/bin/gpasm")
		("command-line" . "%fp -p p16f84 %lft")
		("command-err" . t))))

(defvar ceta-command-file nil
  "Executable full-path file name.")
(defvar ceta-command-line nil
  "Command line including execute options and target/executable file-name-macros.")
(defvar ceta-command-err nil
  "Boolean value which specifies stdout and stderr to be sent to temporary buffer.")
(defvar ceta-command-history nil
  "List of ceta command history.")

(defun ceta-get-command-variables (ceta-command-item-buf)
  "Get and Set basical command information"
  (progn
	  (setq ceta-command-file (cdr (assoc '"command-file" ceta-command-item-buf)))
	  (setq ceta-command-line (cdr (assoc '"command-line" ceta-command-item-buf)))
	  (setq ceta-command-err (cdr (assoc '"command-err" ceta-command-item-buf)))
	  ))

(defun ceta-get-command-completion-alist ()
  "Get alist of ceta-command for using completing-read"
  (let ((list-buf (mapcar 'car ceta-command-alist))
		(alist-buf '())
		(item-num 1))
	(while list-buf
	  (setq alist-buf (cons (cons (car list-buf) item-num) alist-buf))
	  (setq item-num (+ 1 item-num))
	  (setq list-buf (cdr list-buf)))
	(reverse alist-buf)
	))
;
;%fc  : exec file (ex: /usr/X11R6/bin/rxvt or rxvt)
;%lft : target fullpath (ex: /home/hoge/hello.c)
;%lt  : target no path (ex: hello.c)
;
; There are more some unique macros...
;
;%lpt : path of target (ex: /home/hoge/hello.c -> /home/hoge/)
;%ldt : directory of target(long) (ex: /home/hoge/hello.c -> /home/hoge)
;       (NOTE!! Attention to Last SLASH of those two macros)
;%lnt : part of file name except extension (include directry) (long)
;       (ex: /home/hoge/hello.c -> /home/hoge/hello)
;%lot : part of file name except extension and directory (like specification Java class file)
;       (ex: /home/hoge/hello.java -> hello)
(defun ceta-replace-macros (ccf-buf tfn-buf ccl-buf)
  "Replace Ceta Command-Line Macros in temporary buffer."
  (let* ((ceta-fc ccf-buf)
		 (ceta-lft (expand-file-name tfn-buf))
		 (ceta-lt (file-name-nondirectory tfn-buf))
		 (ceta-lpt (file-name-directory tfn-buf))
		 (ceta-ldt (directory-file-name ceta-lpt))
		 (ceta-lnt (file-name-sans-extension tfn-buf))
		 (ceta-lot (file-name-nondirectory ceta-lnt)))
	(replace-string "%fc" ceta-fc nil (point-min) (point-max))
	(replace-string "%lft" ceta-lft nil (point-min) (point-max))
	(replace-string "%lt" ceta-lt nil (point-min) (point-max))
	(replace-string "%lpt" ceta-lpt nil (point-min) (point-max))
	(replace-string "%ldt" ceta-ldt nil (point-min) (point-max))
	(replace-string "%lnt" ceta-lnt nil (point-min) (point-max))
	(replace-string "%lot" ceta-lot nil (point-min) (point-max))
	; Get replaced string
	(buffer-string)
	)
  )

(defun ceta-get-selected-commandset (target-file-name-buf)
  "Get and Set Ceta variables and Return List which consists of
Command Name, Command Line(macros expanded), Command Error Flag."
  (let ((idd-buf insert-default-directory)
		(cic-buf completion-ignore-case)
		(ccm-buf (get-buffer-create "ceta-command-macro-buffer"))
		(curr-buf (current-buffer)))
	(setq insert-default-directory nil)
	(setq completion-ignore-case t)
	(setq ceta-command-name
		  (completing-read
		   "Select Ceta CommandSet: "
		   (ceta-get-command-completion-alist)
		   nil t nil ceta-command-history))
	(setq target-file-name
		  (read-file-name
		   "Target File Name: "
		   nil target-file-name-buf nil target-file-name-buf))
	(ceta-get-command-variables (cdr (assoc ceta-command-name ceta-command-alist)))
	(setq ceta-command-file
		  (read-file-name
		   "Execute File Name: "
		   nil ceta-command-file nil ceta-command-file))
	(setq ceta-command-line
		  (read-string
		   "Edit Command Line Macros: "
		   ceta-command-line nil ceta-command-line))
	; Whether stderr and stdout should be redirected to emacs buffer or not.
	; "DO NOT run in another shell ? " In other words...
	; "Redirect StdErr and StdOut to Emacs Buffer ? "
	; You SHOULD run such a programs like X Window Programs in another shell (so, your answer is "n").
	; Or, you will run gcc or other command-line programs not in another shell and
	; redirect stdout and stderr message to Emacs Buffer, then, your answer is "y".
	; If you want to change "another shell", change "shell-file-name" emacs variable.
	; (like (setq shell-file-name "/usr/local/bin/Eterm"))
	(setq ceta-command-err
		  (y-or-n-p-with-timeout
		   (format "DO NOT run in another shell(%s) ?  " shell-file-name)
										;(In other words, redirect StdErr and StdOut to Emacs Buffer) ?  : "
		   3 ceta-command-err))

	; Replace Command-Line Macros
	; Switch to temporary buffer for replacing macros
	(set-buffer ccm-buf)
	; Insert command line to temporary buffer
	(insert ceta-command-line)
	; Get replaced string
	(setq ceta-command-line-expanded-buf
		  (ceta-replace-macros ceta-command-file
							   target-file-name
							   ceta-command-line))
	; Set temporary buffer unmodified for killing it without confirmation
	(set-buffer-modified-p nil)
	; Restore to Current buffer
	(set-buffer curr-buf)
	; Kill temporary buffer
	(kill-buffer ccm-buf)

	(setq ceta-command-line-expanded
		  (read-string
		   "Last Chance : "
		   ceta-command-line-expanded-buf nil ceta-command-line-expanded-buf))
	(setq completion-ignore-case cic-buf)
	(setq insert-default-directory idd-buf)
	(list ceta-command-name ceta-command-line-expanded ceta-command-err)
	)
  )

(defun ceta-buffer-readonly (cb tb is-ro)
  "Toggle read-only property of tb to is-ro.
cb is stand for Current Buffer, tb is stand for Target Buffer."
  (when tb
	(set-buffer tb)
	(setq buffer-read-only is-ro)
	(set-buffer cb)
	)
  )

(defun ceta-menu (ceta-command-name-buf ceta-command-line-expanded-buf ceta-command-err-buf)
  "Ceta External Generic Shell Service."
  (interactive
   ;; Select Cete command menu from mini-buffer.
   (let ((target-file-name (buffer-file-name)))
	 (ceta-get-selected-commandset target-file-name)
	 ))
  (save-excursion
	(let ((df default-directory))
	  (setq default-directory df)
	  (cd df))
	
	;; Enable buffer-writing Temporarily.
	;; (buffers showing debug messages are usually read-only)
	(ceta-buffer-readonly (current-buffer) (get-buffer "ceta-debug") nil)

	(if ceta-command-err-buf
		(eval (append '(start-process "ceta-command" "ceta-debug")
			  (split-string ceta-command-line-expanded-buf)))
	  (start-process-shell-command "ceta-proc" "ceta-debug"
								   ceta-command-line-expanded-buf)
	  )
	;; Show debug message buffer.
	(unless (get-buffer-window "ceta-debug")
	  ;; If debug buffer was not created, split other window and show debug buffer.
	  (setq nw (split-window (next-window)))
	  (select-window nw)
	  (set-window-buffer nw "ceta-debug")
	  )
	
	;; Disable buffer-writing.
	(ceta-buffer-readonly (current-buffer) (get-buffer "ceta-debug") t)
	)
  )

(require 'bookmark)

(defvar ceta-bookmark-template-p t
  "Whether Ceta Template Feature using Bookmark Mode or not.
If nil, don't use ceta template feature. Or else, use.")

(defun ceta-bookmark-template-toggle (&optional arg)
  "Toggle ceta-bookmark-template-p variable."
  (interactive "P")
  (if (not (null arg))
	  (progn
		(setq ceta-bookmark-template-p arg)
		(message
		 (format "Ceta Template Feature is set to %s" arg))
		)
	(let ((curr-cbt ceta-bookmark-template-p))
	  (setq ceta-bookmark-template-p (not curr-cbt))
	  (message
	   (format "Ceta Template Feature is toggled to %s" ceta-bookmark-template-p)
	   )
	  )
	)
  )

(defadvice bookmark-jump-noselect (after ceta-bookmark-template activate compile)
  "The Chance to change buffer-file-name. 
Depend to how to use, it will work Template feature."
  (save-excursion
	(if (and ceta-bookmark-template-p
			 ad-return-value
			 ;; Confirm whether selected bookmark is not directory
			 ;; or something like info file.
			 (buffer-file-name (car ad-return-value))
			 )
		(let* ((curr-buffer (current-buffer))
			   (cbt-buffer (car ad-return-value))
			   (cbt-file-name (buffer-file-name cbt-buffer))
			   (insert-default-directory nil)
			   (cbt-new-file-name (read-file-name
								   "New File Name : "
								   nil cbt-file-name nil cbt-file-name))
			   )
		  ;; If new file name is same as old, don't change visited-file-name.
		  (message (format "new: %s, \nold: %s"
						   (expand-file-name cbt-new-file-name)
						   (expand-file-name cbt-file-name)))
		  (if (not (equal (expand-file-name cbt-new-file-name)
						  (expand-file-name cbt-file-name)))
			  (progn
				(set-buffer cbt-buffer)
				(set-visited-file-name cbt-new-file-name)
				(set-buffer curr-buffer)
				)
			)
		  )
	  )
	)
  )

(defun khai-command (command buffer)
  "Execute some command on buffer.  Not a official function."
  (save-excursion
    (YaTeX-showup-buffer
     buffer (function (lambda (x) (nth 3 (window-edges x)))))
    (let ((df default-directory))		;preserve current buf's pwd
      (set-buffer (get-buffer-create buffer))	;1.61.3
      (setq default-directory df)
      (cd df))
    (erase-buffer)
    (if (not (fboundp 'start-process))
	(call-process
	 shell-file-name nil buffer nil YaTeX-shell-command-option command)
      (if (and (get-buffer-process buffer)
	       (eq (process-status (get-buffer-process buffer)) 'run)
	      (not
		(y-or-n-p (format "Process %s is running. Continue?" buffer))))
	  nil
	(set-process-buffer
	 (start-process
	  "system" buffer shell-file-name YaTeX-shell-command-option command)
	 (get-buffer buffer))))))
